local merkModelUtil
local merkSettingsUtil
local version = 2
local minorVersion = 1

require "stringutil"

-- base functions
local function fileExists(fileName)
	local file = io.open(fileName, "r")
	if file then
		file:close()
		return true
	end
	return false
end

local next = next
local function tableEmpty(t)
	return next(t) == nil
end

local function tableEqual(t1, t2)
	if type(t1) ~= "table" or type(t2) ~= "table" then
		return false
	end
	local isEqual = true
	local checkedKeys = {}
	for k, v in pairs(t1) do
		if (t2[k] == nil) or (type(t2[k]) ~= type(v)) then
			isEqual = false
			break
		elseif type(v) == "table" then
			isEqual = tableEqual(v, t2[k])
			if not isEqual then
				break
			end
		elseif v ~= t2[k] then
			isEqual = false
			break
		end
		checkedKeys[k] = true
	end
	if isEqual then
		for k, v in pairs(t2) do
			if not checkedKeys[k] then
				if (t1[k] == nil) or (type(t1[k]) ~= type(v)) then
					isEqual = false
					break
				elseif type(v) == "table" then
					isEqual = tableEqual(v, t1[k])
					if not isEqual then
						break
					end
				elseif v ~= t1[k] then
					isEqual = false
					break
				end
			end
		end
	end
	return isEqual
end

local function tableCopy(input)
	if type(input) == "table" then
		local result = {}
		for k, v in pairs(input) do
			if type(v) == "table" then
				result[k] = tableCopy(v)
			else
				result[k] = v
			end
		end
		return result
	else
		return input
	end
end

local function getModPath()
	local info
	local i = 1
	repeat
		info = debug.getinfo(i,'S')
		i = i+1
	until info == nil
	
	local modPath = ""
	local filePath = ""
	if i > 1 then
		info = debug.getinfo(i-2,'S')
		if info then
			filePath = info.source
			if string.ends(info.source, 'mod.lua') then
				modPath = string.gsub(info.source, "@(.*/)mod[.]lua", "%1")
			elseif string.ends(info.source, '.mdl') then
				modPath = string.gsub(info.source, "@(.*/)res/models/model/.+[.]mdl", "%1")
			elseif string.ends(info.source, '.con') then
				modPath = string.gsub(info.source, "@(.*/)res/construction/.+[.]con", "%1")
			elseif string.ends(info.source, '.msh') then
				modPath = string.gsub(info.source, "@(.*/)res/models/mesh/.+[.]msh", "%1")
			elseif string.ends(info.source, '.grp') then
				modPath = string.gsub(info.source, "@(.*/)res/models/group/.+[.]grp", "%1")
			elseif string.ends(info.source, '.mtl') then
				modPath = string.gsub(info.source, "@(.*/)res/models/material/.+[.]mtl", "%1")
			elseif string.ends(info.source, '.ani') then
				modPath = string.gsub(info.source, "@(.*/)res/models/animation/.+[.]ani", "%1")
			elseif string.find(info.source, "/workshop/content/1066780/") then
				modPath = string.gsub(info.source, "@(.*/workshop/content/1066780/[^/]*/).*", "%1")
			elseif string.find(info.source, "/staging_area/") then
				modPath = string.gsub(info.source, "@(.*/staging_area/[^/]*/).*", "%1")
			elseif string.find(info.source, "/mods/") then
				modPath = string.gsub(info.source, "@(.*/mods/[^/]*/).*", "%1")
			end
		end
	end
	
	return modPath, filePath
end

local function clamp(value, min, max)
	if type(value) == "number" then
		return math.min(math.max(value, min), max)
	else
		return 0
	end
end

-- Metadata
local function changeMetadataTable(metadata, key, value, table_add)
	if type(metadata) == "table" then
		if metadata[key] then
			if type(value.value) == "table" then
				local new_table
				if type(metadata[key]) ~= "table" then
					if value.action == "REPLACE" or value.action == "ADD_REPLACE" then
						metadata[key] = {}
						new_table = not value.emptyTable
					else
						return
					end
				end
				for k, v in pairs(value.value) do
					changeMetadataTable(metadata[key], k, v, new_table)
				end
			else
				if value.action == "REPLACE" or value.action == "ADD_REPLACE" then
					metadata[key] = value.value
				elseif value.action == "FACTOR" then
					if type(metadata[key]) == "number" and type(value.value) == "number" then
						metadata[key] = metadata[key] * value.value
					end
				end
			end
		elseif value.action == "ADD_REPLACE" or value.action == "ADD" or table_add then
			if type(value.value) == "table" and not value.emptyTable then
				metadata[key] = {}
				for k, v in pairs(value.value) do
					changeMetadataTable(metadata[key], k, v, table_add)
				end
			else
				metadata[key] = value.value
			end			
		end
	end
end

local function changeMetadata(fileName, data)
	if type(fileName) == "string" and type(merkModelUtil) == "table" then
		local metadata = {}
		if type(merkModelUtil.metadata) == "table" and type(merkModelUtil.metadata[fileName]) == "table" then
			metadata = merkModelUtil.metadata[fileName]
		end

		if not tableEmpty(metadata) and type(data) == "table" and type(data.metadata) == "table" then
			for key, value in pairs(metadata) do
				if key ~= "done" then
					changeMetadataTable(data.metadata, key, value)
				end
			end
		end
	else
		-- wrong parameters
		if type(fileName) ~= "string" then
			print(string.format("[ModUtil|Models] bad argument #1 to 'changeMetadata' (string expected, got %s)", type(fileName)))
		end
	end
end

local actionList = {
	FACTOR = {
		REPLACE = "REPLACE",
		ADD = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	REPLACE = {
		FACTOR = "REPLACE",
		ADD = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	ADD = {
		FACTOR = "ADD_REPLACE",
		REPLACE = "ADD_REPLACE",
		ADD_REPLACE = "ADD_REPLACE",
	},
	ADD_REPLACE = {
		FACTOR = "ADD_REPLACE",
		REPLACE = "ADD_REPLACE",
		ADD = "ADD_REPLACE",
	},
}

local function setMetadataTable(metadataCache, key, value, priority, action, empty)
	if metadataCache[key] == nil or (type(metadataCache[key]) == "table" and type(metadataCache[key].value) ~= "table" and metadataCache[key].priority < priority and not empty) then
		metadataCache[key] = {
			value = {},
			priority = priority,
			action = action,
		}
		if type(value) == "table" then
			metadataCache[key].emptyTable = empty
			for k, v in pairs(value) do
				setMetadataTable(metadataCache[key].value, k, v, priority, action, empty)
			end
		else
			metadataCache[key].value = value
		end
		return action, priority
	elseif type(metadataCache[key]) == "table" and type(metadataCache[key].value) ~= "table" and metadataCache[key].priority >= priority then
		return metadataCache[key].action or "ADD_REPLACE", metadataCache[key].priority
	elseif type(metadataCache[key]) == "table" and type(metadataCache[key].value) == "table" then
		if type(value) == "table" then
			local act, prio
			for k, v in pairs(value) do
				act, prio = setMetadataTable(metadataCache[key].value, k, v, priority, action)
				metadataCache[key].action = actionList[act][metadataCache[key].action or "ADD_REPLACE"]
				metadataCache[key].priority = math.max(prio, metadataCache[key].priority)
			end
			return metadataCache[key].action, metadataCache[key].priority
		else
			if metadataCache[key].priority < priority then
				metadataCache[key] = {
					value = value,
					priority = priority,
					action = action
				}
				return action, priority
			else
				return metadataCache[key].action or "ADD_REPLACE", metadataCache[key].priority
			end
		end
	end
	
	return "ADD_REPLACE", 0
end

local function setMetadata(file_name, meta_data, prio, act)
	if type(merkModelUtil) == "table" then
		local fileName = file_name
		local metadata = meta_data
		local priority = prio or 0
		local action = act or "ADD_REPLACE"
		
		if type(fileName) == "string" and type(metadata) == "table" and type(priority) == "number" and type(action) == "string" then
			if type(merkModelUtil.metadata) ~= "table" then
				merkModelUtil.metadata = {}
			end
			if type(merkModelUtil.metadata[fileName]) ~= "table" then
				merkModelUtil.metadata[fileName] = {}
			end
			
			local metadataCache
			
			if type(metadata) == "table" then
				for key, value in pairs(metadata) do
					if string.find(key, "[.]") then
						local new_value = {}
						local new_key = string.match(key, "([^.]+)[.]")
						local current_value = new_value
						for k in string.gmatch(key, "[.]([^.]+)[.]") do
							current_value[key] = {}
						end
						
						metadataCache = merkModelUtil.metadata[fileName]
						
						setMetadataTable(metadataCache, new_key, new_value, priority, action, true)
						current_value[string.gsub(key, "([^.]+)[.]", "")] = value
						setMetadataTable(metadataCache, new_key, new_value, priority, action)
					else
						metadataCache = merkModelUtil.metadata[fileName]
						
						setMetadataTable(metadataCache, key, value, priority, action)
					end
				end
			end
			
			-- delete empty tables
			if tableEmpty(merkModelUtil.metadata[fileName]) then
				merkModelUtil.metadata[fileName] = nil
				if tableEmpty(merkModelUtil.metadata) then
					merkModelUtil.metadata = nil
				end
			end
		else
			-- wrong parameters
			if type(fileName) ~= "string" then
				print(string.format("[ModUtil|Models] bad argument #1 to 'setMetadata' (string expected, got %s)", type(fileName)))
			end
			if type(metadata) ~= "table" then
				print(string.format("[ModUtil|Models] bad argument #2 to 'setMetadata' (table expected, got %s)", type(metadata)))
			end
			if type(priority) ~= "number" then
				print(string.format("[ModUtil|Models] bad argument #3 to 'setMetadata' (number expected, got %s)", type(priority)))
			end
			if type(action) ~= "string" then
				print(string.format("[ModUtil|Models] bad argument #4 to 'setMetadata' (string expected, got %s)", type(action)))
			end
		end
	end
end

-- easy repaint creation, no changes inside .mdl necessary
local function makeRepaint(fileName, materials, optionals)
	if type(fileName) == "string" and type(materials) == "table" then
		if type(merkModelUtil.repaintList) ~= "table" then
			merkModelUtil.repaintList = {}
		end
		
		-- check material list format
		local mat = {}
		for k, v in pairs(materials) do
			if type(k) ~= "string" then
				print(string.format("[ModUtil|Models] bad argument #2 (key) to 'makeRepaint' (string expected, got %s)", type(k)))
			elseif type(v) ~= "string" then
				print(string.format("[ModUtil|Models] bad argument #2 (value) to 'makeRepaint' (string expected, got %s)", type(v)))
			else
				mat[k] = v
			end
		end
		
		local opt = optionals or {}
		local base, meta, files = opt.base, opt.metadata
		
		-- check optional parameters
		if base then
			if type(base) ~= "string" then
				print(string.format("[ModUtil|Models] bad argument #3.base to 'makeRepaint' (string expected, got %s)", type(base)))
				base = nil
			else
				if type(merkModelUtil.repaintBase) ~= "table" then
					merkModelUtil.repaintBase = {}
				end
				
				-- tell modifier to save this models data
				if not merkModelUtil.repaintBase[base] then
					merkModelUtil.repaintBase[base] = 0
				end
				merkModelUtil.repaintBase[base] = merkModelUtil.repaintBase[base] + 1
			end
		end
		
		if meta then
			if type(meta) ~= "table" then
				print(string.format("[ModUtil|Models] bad argument #3.metadata to 'makeRepaint' (table expected, got %s)", type(meta)))
				meta = nil
			else
				-- hand over metadata changes to the modifier
				setMetadata(fileName, meta)
			end
		end
		
		if opt.files then
			if type(opt.files) ~= "table" then
				print(string.format("[ModUtil|Models] bad argument #3.files to 'makeRepaint' (table expected, got %s)", type(meta)))
			else
				-- switch from plain list to key-value pairs (to simulate "in" operator)
				files = {}
				for f, file in ipairs(opt.files) do
					files[file] = true
				end
			end
		end
		
		-- save repaint data
		merkModelUtil.repaintList[fileName] = {
			matReplace = mat,
			base = base,
			files = files,
		}
	else
		if type(fileName) ~= "string" then
			print(string.format("[ModUtil|Models] bad argument #1 to 'makeRepaint' (string expected, got %s)", type(fileName)))
		end
		if type(materials) ~= "table" then
			print(string.format("[ModUtil|Models] bad argument #2 to 'makeRepaint' (table expected, got %s)", type(materials)))
		end
	end
end

local function replaceMaterials(children, mat, files)
	for c, child in ipairs(children) do
		if type(child) == "table" then
			if child.materials then
				if (not files) or (child.mesh and files[child.mesh]) then		-- only apply to given meshes
					local newmats = {}
					
					-- find materials to replace
					for m, material in ipairs(child.materials) do
						if type(mat[material]) == "string" then
							newmats[m] = mat[material]
						end
					end
					
					-- replace matches
					for k, v in pairs(newmats) do
						child.materials[k] = v
					end
				end
			elseif type(child.children) == "table" then
				-- next node level
				replaceMaterials(child.children, mat, files)
			end
		end
	end
end

local function applyRepaint(data, repaintData)
	-- check if base data is available
	if repaintData.base and type(merkModelUtil.repaintBase) == "table" and type(merkModelUtil.repaintBase[repaintData.base]) == "table" then
		local d = tableCopy(merkModelUtil.repaintBase[repaintData.base].data)
		
		-- remove old data
		for k, v in pairs(data) do
			data[k] = nil
		end
		
		-- insert new data
		for k, v in pairs(d) do
			data[k] = v
		end
		
		-- garbage collection
		merkModelUtil.repaintBase[repaintData.base].users = merkModelUtil.repaintBase[repaintData.base].users - 1
		if merkModelUtil.repaintBase[repaintData.base].users == 0 then
			merkModelUtil.repaintBase[repaintData.base] = nil
		end
	end
	
	-- check data structure
	if type(data) == "table" and type(data.lods) == "table" then
		for l, lod in ipairs(data.lods) do
			if type(lod.node) == "table" and type(lod.node.children) == "table" then
				-- actual material replacement function
				replaceMaterials(lod.node.children, repaintData.matReplace, repaintData.files)
			end
		end
	end
end

local function modelModifierFunction(fileName, data)
	if type(merkModelUtil) == "table" then

		if data and data.metadata then
			local meta = data.metadata
			
			-- insert pre-defined camera configs
			if type(meta.cameraConfig) == "string" then
				if type(merkModelUtil.cameraConfigs) == "table" and merkModelUtil.cameraConfigs[meta.cameraConfig] then
					meta.cameraConfig = tableCopy(merkModelUtil.cameraConfigs[meta.cameraConfig])
				end
			end
		end
		
		local pattern = "models/model"
		local i = string.find(fileName, pattern)
		if i then
			local modelId = string.sub(fileName, i+#pattern+1)
			
			if type(data) == "table" then
				-- save this models data, if it is the base for a repaint
				if type(merkModelUtil.repaintBase) == "table" then
					if type(merkModelUtil.repaintBase[modelId]) == "number" and data.version == 1 then
						merkModelUtil.repaintBase[modelId] = {
							data = tableCopy(data),
							users = merkModelUtil.repaintBase[modelId],
						}
					end
				end
				
				-- apply repaints
				if type(merkModelUtil.repaintList) == "table" then
					if merkModelUtil.repaintList[modelId] then
						applyRepaint(data, merkModelUtil.repaintList[modelId])
					end
				end
				
				-- change metadata
				if type(merkModelUtil.metadata) == "table" then
					changeMetadata(modelId, data)
				end
			end
			
			-- set model availibity for "getModels"
			if type(merkModelUtil.modelList) ~= "table" then
				merkModelUtil.modelList = {}
			end
			merkModelUtil.modelList[modelId] = 2
		end
	end
	
	return data
end

-- wrapper function so that always the latest modifier version is used
local function modelModifier(fileName, data)
	
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].modelModifierFunction(fileName, data)
	end
	
	return data
end

-- check if given model-file is available or return all available models from a given list
local function getModels(modelIDs, availability)
	if type(merkModelUtil.modelList) == "table" then
		if type(modelIDs) == "string" then
			return merkModelUtil.modelList[modelIDs] or 0
		elseif type(modelIDs) == "table" then
			local models = {}
			for __, model in ipairs(modelIDs) do
				if merkModelUtil.modelList[model] >= availability then
					models[#models+1] = model
				end
			end
			return models
		else
			print(string.format("[ModUtil|Models] bad argument #1 to 'getModels' (string or table expected, got %s)", type(modelIDs)))
		end
	end
	
	return modelIDs
end

-- define a camera config, that will be inserted, if the given string is found as the cameraConfig value
local function setCameraConfig(key, config)
	if type(merkModelUtil.cameraConfigs) ~= "table" then
		merkModelUtil.cameraConfigs = {}
	end
	
	if type(key) == "string" and type(config) == "table" then
		if config.positions then
			merkModelUtil.cameraConfigs[key] = config
		else
			merkModelUtil.cameraConfigs[key] = {positions = config}
		end
	else
		-- wrong parameters
		if type(key) ~= "string" then
			print(string.format("[ModUtil|Models] bad argument #1 to 'setCameraConfig' (string expected, got %s)", type(key)))
		end
		if type(config) ~= "table" then
			print(string.format("[ModUtil|Models] bad argument #2 to 'setCameraConfig' (table expected, got %s)", type(config)))
		end
	end
end

-- Object filter
local function modelFilter(fileName, data)
	local pattern = "models/model"

	if type(merkModelUtil) == "table" then
		local i = string.find(fileName, pattern)
		if i then
			local modelId = string.sub(fileName, i+#pattern+1)
			
			if type(merkModelUtil.modelList) ~= "table" then
				merkModelUtil.modelList = {}
			end
			merkModelUtil.modelList[modelId] = 1
			
			if type(merkModelUtil.hiddenObjects) == "table" and merkModelUtil.hiddenObjects[modelId] then
				return false
			end
		end
	end
	
	return true
end

local function objectFilter(fileName, data, pattern)
	if type(merkModelUtil) == "table" and type(merkModelUtil.hiddenObjects) == "table" then
		local i = string.find(fileName, pattern)
		if i then
			local objectId = string.sub(fileName, i+#pattern+1)
			if merkModelUtil.hiddenObjects[objectId] then
				return false
			end
		end
	end
	
	return true
end

local function filterModels(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].modelFilter(fileName, data)
	end
	
	return true
end

local function filterConstructions(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "construction/")
	end
	
	return true
end

local function filterBridges(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/bridge/")
	end
	
	return true
end

local function filterTunnels(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/tunnel/")
	end
	
	return true
end

local function filterStreets(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/street/")
	end
	
	return true
end

local function filterTracks(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/track/")
	end
	
	return true
end

local function filterMultipleUnits(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/multiple_units/")
	end
	
	return true
end

local function filterRailroadCrossings(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/railroad_crossing/")
	end
	
	return true
end

local function filterGroundTextures(fileName, data)
	if merk_modelUtil and merk_modelUtil[version] and merk_modelUtil[version].minorVersion >= minorVersion then
		return merk_modelUtil[version].objectFilter(fileName, data, "config/auto_ground_tex")
	end
	
	return true
end

local function hideObject(objectId)
	if type(objectId) == "string" and type(merkModelUtil) == "table" then
		if type(merkModelUtil.hiddenObjects) ~= "table" then
			merkModelUtil.hiddenObjects = {}
		end
		
		merkModelUtil.hiddenObjects[objectId] = true
	else
		-- wrong parameters
		if type(objectId) ~= "string" then
			print(string.format("[ModUtil|Models] bad argument #1 to 'hideObject' (string expected, got %s)", type(objectId)))
		end
	end
end

local function createEnvironment(options)
	if game then
		if type(game.merkModelUtil) ~= "table" then
			game.merkModelUtil = {}
		end
		if type(game.merkModelUtil[version]) ~= "table" then
			game.merkModelUtil[version] = {}
		end
		merkModelUtil = game.merkModelUtil[version]
		
		return {
			setMetadata = setMetadata,
			changeMetadata = changeMetadata,
		}
	else
		print(string.format("[ModUtil|Models] Creation of working environment failed. It seems you're trying to create the environment outside the \"runFn\"."))
	end
end

local function initialiseModelUtil(addFilters)
	if game then
		if type(game.merkModelUtil) ~= "table" then
			game.merkModelUtil = {}
		end
		if type(game.merkModelUtil[version]) ~= "table" then
			game.merkModelUtil[version] = {}
		end
		merkModelUtil = game.merkModelUtil[version]
	end
	
	
	if addFilters ~= false then
		if addFileFilter then
			addFileFilter("model/vehicle", filterModels)
			addFileFilter("model/tree", filterModels)
			addFileFilter("model/other", filterModels)
			
			addFileFilter("construction", filterConstructions)
			addFileFilter("bridge", filterBridges)
			addFileFilter("tunnel", filterTunnels)
			addFileFilter("street", filterStreets)
			addFileFilter("track", filterTracks)		
			addFileFilter("multipleUnit", filterMultipleUnits)
			addFileFilter("railroadCrossing", filterRailroadCrossings)
			addFileFilter("autoGroundTex", filterGroundTextures)
		end
		if addModifier then
			addModifier("loadModel", modelModifier)
		end
	end

	merk_modelUtil[version] = {
		createEnvironment = createEnvironment,
		setMetadata = setMetadata,
		changeMetadata = changeMetadata,
		hideObject = hideObject,
		minorVersion = minorVersion,
		modelFilter = modelFilter,
		objectFilter = objectFilter,
		modelModifierFunction = modelModifierFunction,
		getModels = getModels,
		setCameraConfig = setCameraConfig,
		makeRepaint = makeRepaint,
	}
end

if not merk_modelUtil or not merk_modelUtil[version] then

	if type(merk_modelUtil) ~= "table" then
		merk_modelUtil = {}
	end
	
	initialiseModelUtil()
end
if merk_modelUtil[version].minorVersion < minorVersion then
	initialiseModelUtil(false)
end

-- settings
local function readUserSettings(id)
	if type(merkSettingsUtil) == "table" and type(merkSettingsUtil[id]) == "table" then
		local settings = merkSettingsUtil[id].values
		
		local modPath = getModPath()
		
		if fileExists(modPath.."mod.lua") and fileExists(modPath.."settings.lua") then
			local settingsFile = loadfile(modPath.."settings.lua")
			local loadedSettings = {}
			
			if type(settingsFile) == "function" then
				loadedSettings = settingsFile()
			else
				print(string.format("[ModUtil|UserSettings] Could not read settings file %q.", modPath.."settings.lua"))
			end
			
			if type(loadedSettings) == "table" then
				for option_key, option in pairs(merkSettingsUtil[id].definitions) do
					if type(loadedSettings[option_key]) == option.type then
						local value = loadedSettings[option_key]
						if option.type == "number" then
							if option.min and value < option.min then
								value = option.min
							end
							if option.max and value > option.max then
								value = option.max
							end
						end
						settings[option_key] = value
					end
				end
			else
				print(string.format("[ModUtil|UserSettings] Settings file %q contains invalid data.", modPath.."settings.lua"))
			end
		end
	end
end

local function createUserSettings(id, options)
	if type(id) == "string" and type(options) == "table" then
		if type(merkSettingsUtil) == "table" then
			if merkSettingsUtil[id] then
				print(string.format("[ModUtil|UserSettings] Previous settings for id %q are overwritten.", id))
			end
			
			merkSettingsUtil[id] = {
				definitions = options,
				values = {},
			}
			
			for option_key, option in pairs(options) do
				if option.type and option.name then
					if option.default ~= nil then
						merkSettingsUtil[id].values[option_key] = option.default
					else
						if option.type == "boolean" then
							merkSettingsUtil[id].values[option_key] = false
						elseif option.type == "number" then
							merkSettingsUtil[id].values[option_key] = 0
						elseif option.type == "string" then
							merkSettingsUtil[id].values[option_key] = ""
						elseif option.type == "table" then
							merkSettingsUtil[id].values[option_key] = {}
						end
					end
				end
			end
			
			readUserSettings(id)
		end
	else
		-- wrong parameters
		if type(id) ~= "string" then
			print(string.format("[ModUtil|UserSettings] bad argument #1 to 'createUserSettings' (string expected, got %s)", type(id)))
		end
		if type(options) ~= "table" then
			print(string.format("[ModUtil|UserSettings] bad argument #2 to 'createUserSettings' (table expected, got %s)", type(options)))
		end
	end
end

local function getUserSettings(id)
	local settings = {}
	
	if type(id) == "string" then
		if type(merkSettingsUtil) == "table" and type(merkSettingsUtil[id]) == "table" then
			settings = merkSettingsUtil[id].values
		end
	else
		-- wrong parameter
		if type(id) ~= "string" then
			print(string.format("[ModUtil|UserSettings] bad argument #1 to 'getUserSettings' (string expected, got %s)", type(id)))
		end
	end
	
	return settings
end

local function initialiseSettingsUtil()
	if game then
		if type(game.merkSettingsUtil) ~= "table" then
			game.merkSettingsUtil = {}
		end
		if type(game.merkSettingsUtil[version]) ~= "table" then
			game.merkSettingsUtil[version] = {}
		end
		merkSettingsUtil = game.merkSettingsUtil[version]
	end

	merk_SettingsUtil[version] = {
		create = createUserSettings,
		get = getUserSettings,
		minorVersion = minorVersion,
	}
end

if not merk_SettingsUtil or not merk_SettingsUtil[version] then
	if type(merk_SettingsUtil) ~= "table" then
		merk_SettingsUtil = {}
	end

	initialiseSettingsUtil()
else
	if merk_SettingsUtil[version].minorVersion < minorVersion then
		initialiseSettingsUtil()
	end
end

-- return value
local models = tableCopy(merk_modelUtil[version])
models.objectFilter = nil
models.modelFilter = nil
models.modelModifierFunction = nil

local mod_util = {
	models = models,
	userSettings = tableCopy(merk_SettingsUtil[version]),
	initialize = function (id, settings, options)
			merk_SettingsUtil[version].create(id, settings)
			return merk_modelUtil[version].createEnvironment(options)
		end,
	minorVersion = minorVersion,
	misc = {
		getModPath = getModPath,
		clamp = clamp,
	},
}

if not merk_modutil or not merk_modutil[version] then
	if type(merk_modutil) ~= "table" then
		merk_modutil = {}
	end
	merk_modutil[version] = mod_util
else
	if merk_modutil[version].minorVersion < minorVersion then
		merk_modutil[version] = mod_util
		package.loaded["merk_modutil_2"] = nil
	end
end

return merk_modutil[version]
